/*
   Boat hull toolbox - a 3d carving method

   Author: Guillaume Fougnies   (sailboat@w0.yt)
   License: GNU GPL
   URL: https://www.thingiverse.com/thing:3707731


    WALLY TANGO 1:25
    study


   Copyright (C) 2020  Guillaume Fougnies

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

 */

processing	= 4;		/* Processing:
				   - 0 : raw solid hull  (for design or molding)
				   - 1 : rounded solid hull
				   - 2 : rounded empty hull
				   - 3 : bulkheads (for splicing bulkheads)
				   - 4 : one bulkhead for print
				   - 5 : deck
				   - 6 : waterlines
				   - 7 : dual hull for design
				   - 8 : lateral hull + waterlines well aligned for design
				 */


phi_steps	= 1.00;		/* precision (phi step increment in range 0-90 )	*/
filled		= 1;		/* filled hull (0/1) - only for processing:0		*/
rounded		= 0;		/* rounded hull - 0 to radius (cpu intensive)		*/
//length		= 30480 -rounded;	/* hull: length  					*/
//width		= 7200;		/* hull: width   					*/
//height		= 3100;		/* hull: height - for a circle, set it to: (width/2) 	*/

length		= (30480)/25 -rounded;	/* hull: length  					*/
width		= (7200)/25;		/* hull: width   					*/
height		= (3100)/25;		/* hull: height - for a circle, set it to: (width/2) 	*/




echo( length );
echo( width );
echo( height );

hull_thickness		= 3;    /* for empty hull rendering */
bulkhead_thickness	= 4;
no_stem			= 1;	/* disable stem */
mesh_thickness		= 9;

waterlines		= 10;
waterline_rotate	= 0;	/* rotation angle of the hull for waterlines */

spacer			= 0.3;

$fn			= 50;

XPS_thickness		= 40;
BH_thickness		= 3;

BH_id			= 0; /* Processing: 4 -> the BH to show */

BH_xpsnb = [
	0,
	1,
	2,
	3,
	3,
	3,
	3,
	3,
	3,
	3,
	3,
	2.6
];



/*
   The shape:
 
   - step: 1 .. 100 - from 0 to 'length'. (1 and 100 are mandatory)
   - height_factor / width_factor:
             minimum - 0.01 .. 1.00 - maximum 'height' / 'width'
   - shape_factor:
             0.00 .. 1.00 - circle to ridge
 
 */
shape = [
        // step , [ height_start_deck_factor, height_factor, width_factor, shape_factor, depth_factor ]
	[1,   [ 0.02, 	0.765, 	0.020, -0.015, 0.60 ]],
	[2,   [ 0.01,	0.805, 	0.045, -0.015, 0.60 ]],
	[3,   [ 0.005,	0.82, 	0.065, -0.015, 0.60 ]],
	[6,   [ 0,	0.845, 	0.13, -0.015 , 0.60 ]],
	[9,   [ 0.005,	0.865, 	0.20, -0.015, 0.60 ]],
	[12,  [ 0.01,	0.885, 	0.26, -0.020, 0.60 ]],
	[15,  [ 0.015,	0.905, 	0.32, -0.025 , 0.65 ]],
	[18,  [ 0.02,	0.92, 	0.38, -0.03 , 0.70 ]],
	[21,  [ 0.025,	0.93, 	0.44, -0.03, 0.75 ]],
	[24,  [ 0.027,	0.94, 	0.505, -0.03, 0.80 ]],
	[28,  [ 0.028,	0.95, 	0.59, -0.030 , 0.85 ]],
	[33,  [ 0.029,	0.97, 	0.69, -0.043 , 0.88 ]],
	[40,  [ 0.031,	0.985,	0.81, -0.043 , 0.88 ]],
	[45,  [ 0.033,	0.990,	0.879,  -0.053 , 0.88 ]],
	[50,  [ 0.040,	0.993,	0.929,  -0.053 , 0.88 ]],
	[55,  [ 0.051,	0.992,	0.965, -0.076, 0.88 ]],
	[60,  [ 0.06,	0.977,	0.988, -0.068 , 0.88 ]],
	[65,  [ 0.070,	0.958,	1.00, -0.080 ,  0.88 ]],
	[75,  [ 0.095,	0.905,	1.00, -0.090 , 0.84 ]],
	[80,  [ 0.11,	0.874,	0.995, -0.100, 0.80 ]],
	[85,  [ 0.14,	0.837,	0.990, -0.111, 0.75 ]],
	[90,  [ 0.16,	0.800,	0.985, -0.124, 0.70 ]],
	[95,  [ 0.185,	0.76,	0.980, -0.133, 0.65 ]],
	[100, [ 0.215,	0.72,	0.975, -0.14,  0.60 ]]
];

/*
   Stem:
   step to split - or empty if no split
 */
stemsplit = [ 30, 65 ];


/*
   Helper for 'do.sh'
 */

if (processing == 66) {
	/* execute order 66 */
	echo( len(shape) );
}
if (processing == 67) {
	/* execute order 67 */
	echo( len(stemsplit) );
}


/*
   Computing
 */

bhid = 23;
stid = 0;

if (processing == 2) { filled = 0; }

function mcorrectXcos( phi, roundratio )	= cos(phi/(1+roundratio));
function mcorrectXcospow( phi, depthratio )	= 0;

function mcorrectXdirect( radius, phi, roundratio, depthratio ) =
	radius *
	((phi>90 && phi<180)?
	 (
	  -1 *
	  (abs(mcorrectXcos(phi,roundratio)) + abs(mcorrectXcospow(phi,depthratio)))/
	  (abs(mcorrectXcos(180,roundratio)) + abs(mcorrectXcospow(180,depthratio)))
	 )
	 :
	 (mcorrectXcos( phi, roundratio ))
	);
function mcorrectXdirect( radius, phi, roundratio, depthratio ) = radius * mcorrectXcos( phi, roundratio );

function mcorrectY( radius, phi, fy, roundratio, depthratio ) =
	fy * radius * sin(phi/(1+roundratio));

function mcorrectYsinpow( phi, depthratio ) = pow(sin(phi), 1.00 - depthratio);

function mcorrectY( radius, phi, fy, roundratio, depthratio ) =
	fy * radius *
	((phi>90 && phi<180)?
	 (
	  (sin(phi/(1+roundratio)) + mcorrectYsinpow(phi, depthratio)) /
	  (sin(90/(1+roundratio)) + mcorrectYsinpow(90, depthratio))
	 )
	 :
	 sin(phi/(1+roundratio)));

function mcorrectX( radius, phi, fx, roundratio, depthratio ) =
	fx * (
			mcorrectXdirect( radius, phi, roundratio, depthratio )      /* the curve */
			- mcorrectXdirect( radius, 90, roundratio )     /* regulate position */
	     );

function mcorrectXnormalized( radius, phi, fx, roundratio, depthratio ) =
	mcorrectX( radius, phi, fx, roundratio, depthratio )
	/* Normalize depending on the roundratio */
	* (mcorrectX( radius, 180, fx, 1, 0 ) / mcorrectX( radius, 180, fx, roundratio, 0 ))
	/* Normalize depending on the resulting height */
	* (height / mcorrectX( height, 180, 1, 1, 0 ))
	;

function mcorrectXmax( r_x, fx, roundratio, depthratio ) = [
	mcorrectXnormalized( r_x, 180, fx, roundratio, depthratio )
];
function mcorrectXratio( r_x, phi, fx, roundratio ) = [
	mcorrectXmax( r_x, fx, roundratio )
];

function mcorrect(r_x, r_y, fx, fy, roundratio, depthratio) = [
	for (phi=[(filled?0:90) : phi_steps : 180]) [
		mcorrectXnormalized( r_x, phi, fx, roundratio, depthratio ),
		mcorrectY( r_y, phi, fy, 0, depthratio )
	]
];

function mcorrectInv(r_x, r_y, fx, fy, roundratio, depthratio) = [
	for (phi=[(filled?0:90) : phi_steps : 180]) [
		mcorrectY( r_x, phi, fx, 0, depthratio ),
		mcorrectXnormalized( r_y, phi, fy, roundratio, depthratio )
	]
];

/* a step - w/ targeted: h / w */
module _cockpit_step( step, h, w ) {
	translate([ step[0]*h, 0, 0 ]) {
		difference() {
			linear_extrude(height=0.01) {
				if (step[3] < 0.0) {
					polygon(mcorrectInv( h, w, step[1] - step[0],
								step[2],
								step[3] * -1.0,
								step[4] ));
				}
				else {
					polygon(mcorrect( h, w, step[1] - step[0],
								step[2],
								step[3],
								step[4] ));
				}
			}
			translate([-width*5,0,0])
				cube([width*10,width*10,10], center=true);
		}
	}
}

/* a step */
module cockpit_step( step ) {
	_cockpit_step( step, height, width/2 );
}

/* a step bulkhead */
module cockpit_step_bh( step ) {
	difference() {
		cockpit_step( step );
		/* scale/translate to make openscad preview happy */
		translate([0,0,-0.1]) scale([1,1,100])
			_cockpit_step( step, height*.7, (width/2)*.7 );
	}
}

/* a slice between two steps */
module cockpit_slice( b ) {
	for (p=[b:1:b+1]) {
		/* Move it to z axis */
		translate([0,0,((shape[p][0]-1)/99)*length]) {
			cockpit_step( shape[p][1] );
		}
	}
}

/* half of the hull */
module cockpit_half() {
	for (b=[0:1:len(shape)-2]) {
		hull() {
			cockpit_slice( b );
		}
	}
}

/* full hull */
module cockpit() {
	union() {
		cockpit_half();
		mirror([0,1,0]) cockpit_half();
	}
}

/* rounded (minkowski) hull */
module cockpit_rounded() {
	if (rounded)
		minkowski() {
			cockpit();
			sphere(r=rounded);
		}
	else
		cockpit();
}

function sc_reductor( sz, wall ) = ((sz - wall*2)/sz);

module cockpit_mask( wall ) {
	translate( [-0.02, 0, wall] )   /* 0.02 for openscad preview */
		scale([ sc_reductor( height, wall/2 ),
				sc_reductor( width, wall ),
				sc_reductor( length, wall ) ])
		cockpit_rounded();
}

module cockpit_empty( wall ) {
	difference() {
		cockpit_rounded();
		cockpit_mask( wall );
	}
}


module waterline_mask( ) {
	translate([1,-width*2,-length*2])
		cube([(height/waterlines)-0.1,width*4,length*4]);
}

module bh_mask( z, thickness ) {
	translate([-height*2,-width*2, z + thickness/2]) {
		cube([height*4,width*4,length+1]);
		translate([0, 0, -length - 1 - thickness ]) {
			cube([height*4,width*4,length+1]);
		}
	}
}

module bh_plate( z, thickness ) {
	translate([-height*2,-width*2, z - thickness/2]) {
		cube([height*4,width*4,thickness]);
	}
}

function XPS_nb(pos)	= BH_xpsnb[pos] + ((pos>0)?XPS_nb(pos-1):0);
function XPS_sz(pos)	= XPS_nb( pos ) * XPS_thickness + pos * BH_thickness + BH_thickness / 2;

module bhe( x ) {
	bh_plate( XPS_sz(x), BH_thickness );
}

module cockpit_bh() {
	rotate([0,90,0])
	intersection() {
		cockpit();
		union() {
			for (b=[0:1:len( BH_xpsnb )-1]) {
				bhe( b );
			}
		}
	}
}

module bh( x ) {
	rotate([0,0,90])
	intersection() {
		cockpit();
		bhe( x );
	}

}

/* Generate */

//	$vpt=[0,-130,0];
//	$vpr=[0,0,0];
//	$vpd=794; /* move until the cube is 400mm large on screen */

/* Generate */
if (processing == 0) {
	rotate([0,90,0])
		cockpit();
}
else if (processing == 1) {
	rotate([0,90,0])
		cockpit_rounded();
}
else if (processing == 2) {
	rotate([0,90,0])
		cockpit_empty( hull_thickness );
}
else if (processing == 3) {
	cockpit_bh();
	/*
	translate([0,0,height])
		rotate([0,90,0])
		cockpit(); */

}
else if (processing == 4) {
	sca=1.0;

	rotate([0,0,180])
	scale([sca,sca,sca])
	{
		translate([0,150 - 20,-20])
			cube([400,300,1],center=true);
		color([0,0,0])
			translate([0,0,- XPS_sz( BH_id ) ])
			bh( BH_id );
	}
}
//////////////////	rotate([0,90,0]) bh_cockpit_mesh(1);
//	cockpit_step( [ 0.0,  0.97,   0.560, 0.0, 0.4 ] );
else if (processing == 6) {
	rotate([0,90,0])
	difference()
	{
		cockpit();
		for (x=[0 : height/waterlines : height]) {
			translate([x,0,0])
				waterline_mask();
		}
	}
}
else if (processing == 7) {
	translate([0,0,width/2])
		rotate([270,0,0])
		rotate([0,90,0])
		cockpit();
	rotate([0,90,0])
		cockpit();
}
else if (processing == 8) {
	translate([0,0,width/2])
		rotate([270,0,0])
		rotate([0,90,0])
		if (waterline_rotate) {
			difference()
			{
				rotate([0,0,waterline_rotate])
					cockpit();
				for (x=[0 : height/(2*waterlines) : height*2]) {
					translate([x*2 - height/2,0,0])
						waterline_mask();
				}
			}
		}
		else {
			difference()
			{
				cockpit();
				for (x=[0 : height/waterlines : height]) {
					translate([x,0,0])
						waterline_mask();
				}
			}
		}
	rotate([0,90,0])
		cockpit();
}

